@@ -1,14 +1,35 @@
-import socket, struct, os
+import os
+import socket
+import struct
+
+
+class AutoscoperServerError(Exception):
+    """Exception raised when the server reports an error."""
+
+    def __str__(self):
+        return f"Autoscoper Server error: {super().__str__()}"
+
+
+class AutoscoperConnectionError(Exception):
+    """Exception raised when the connection to the server is lost."""
+
+    def __str__(self):
+        return f"Error communicating with Autoscoper server: {super().__str__()}"
 
 
 class AutoscoperConnection:
     def __init__(self, address="127.0.0.1", verbose=False) -> None:
         self.address = address
         self.verbose = verbose
-        self.socket = self.openConnection()
-        self.is_connected = self.test_connection()
+        self.socket = self._openConnection()
 
-    def wait_for_server(self):
+    def __str__(self):
+        return f"Autoscoper connection to {self.address}"
+
+    def __repr__(self):
+        return f"AutoscoperConnection('{self.address}', verbose={self.verbose})"
+
+    def _wait_for_server(self):
         """
         Internal function, should not be called by a user.
 
@@ -19,7 +40,46 @@ class AutoscoperConnection:
             if data:
                 return data
 
-    def test_connection(self):
+    def _pack_data(self, *args):
+        packed_data = bytearray()
+        for arg in args:
+            if isinstance(arg, int):
+                packed_data.extend(arg.to_bytes(4, byteorder="little", signed=False))
+            elif isinstance(arg, float):
+                packed_data.extend(struct.pack("d", arg))
+            elif isinstance(arg, str):
+                packed_data.extend(arg.encode("utf-8"))
+            else:
+                raise Exception(f"Invalid argument type: {type(arg)}")
+        return packed_data
+
+    def _send_command(self, command, *args):
+        """
+        Send a command to the server and wait for the response.
+
+        :param command: The command byte to send
+        :type command: int
+        :param args: The arguments to send with the command
+        :type args: tuple
+        :return: The response from the server
+        :rtype: bytes
+        :raises Exception: If the server response is invalid
+        """
+        packed_data = self._pack_data(*args)
+        b = bytearray()
+        b.append(command)
+        b.extend(packed_data)
+        try:
+            self.socket.sendall(b)
+            response = self._wait_for_server()
+        except socket.error as e:
+            raise AutoscoperConnectionError(e)
+        if response[0] != command:
+            self.closeConnection()
+            raise AutoscoperServerError(f"received {response[0]}, expected {command}")
+        return response
+
+    def _test_connection(self):
         """
         Test the connection to the PyAutoscoper server.
 
@@ -28,16 +88,10 @@ class AutoscoperConnection:
         """
         if self.verbose:
             print("Testing connection")
-        b = bytearray()
-        b.append(0x00)
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x00:
-            self.closeConnection()
-            raise Exception("Server Error testing connection")
+        self._send_command(0x00)
         return True
 
-    def openConnection(self):
+    def _openConnection(self):
         """
         Internal function, should not be called by a user.
 
@@ -49,27 +103,33 @@ class AutoscoperConnection:
         s.connect((self.address, 30007))
         return s
 
+    @property
+    def is_connected(self):
+        """
+        Returns the status of the connection to the PyAutoscoper server.
+
+        :rtype: Boolean
+        """
+        try:
+            self._test_connection()
+            return True
+        except (AutoscoperServerError, AutoscoperConnectionError):
+            return False
+
     def loadTrial(self, trial_file):
         """
         Load a trial file into the PyAutoscoper server.
 
         :param trial_file: The path to the trial file to load
         :type trial_file: str
-        :raises Exception: If the trial file is not found, or If the server fails to load the trial file
+        :raises AutoScoperServerError: If the server fails to load the trial file
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Loading trial file: {trial_file}")
         if not os.path.exists(trial_file):
-            self.closeConnection()
-            raise Exception("Trial file not found")
-        b = bytearray()
-        b.append(0x01)
-        b.extend(trial_file.encode("utf-8"))
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x01:
-            self.closeConnection()
-            raise Exception("Server Error loading trial file")
+            raise AutoscoperServerError(f"File not found: {trial_file}")
+        self._send_command(0x01, trial_file)
 
     def loadTrackingData(
         self,
@@ -101,28 +161,24 @@ class AutoscoperConnection:
         :type is_rad: bool
         :param interpolate: Optional - If true, the tracking data will be interpolated using the spline method. If false, the tracking data will be saved as is (with NaN values). Defaults to false.
         :type interpolate: bool
-        :raises Exception: If the tracking data file is not found, or If the server fails to load the tracking data
+        :raises AutoScoperServerError: If the server fails to load the tracking data
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Loading tracking data: {tracking_data}")
         if not os.path.exists(tracking_data):
-            self.closeConnection()
-            raise Exception("Tracking data file not found")
-        b = bytearray()
-        b.append(0x02)
-        b.extend(volume.to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(is_matrix)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(is_rows)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(is_with_commas)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(is_cm)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(is_rad)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(interpolate)).to_bytes(4, byteorder="little", signed=False))
-        b.extend(tracking_data.encode("utf-8"))
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x02:
-            self.closeConnection()
-            raise Exception("Server Error loading tracking data")
+            raise AutoscoperServerError(f"Tracking data not found: {tracking_data}")
+        self._send_command(
+            0x02,
+            volume,
+            int(is_matrix),
+            int(is_rows),
+            int(is_with_commas),
+            int(is_cm),
+            int(is_rad),
+            int(interpolate),
+            tracking_data,
+        )
 
     def saveTracking(
         self,
@@ -154,25 +210,23 @@ class AutoscoperConnection:
         :type convert_to_rad: bool
         :param interpolate: Optional - If true, the tracking data will be interpolated using the spline method. If false, the tracking data will be saved as is (with NaN values). Defaults to false.
         :type interpolate: bool
-        :raises Exception: If the server fails to save the tracking data
+        :raises AutoScoperServerError: If the server fails to save the tracking data
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Saving tracking data: {tracking_file}")
-        b = bytearray()
-        b.append(0x03)
-        b.extend(volume.to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(save_as_matrix)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(save_as_rows)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(save_with_commas)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(convert_to_cm)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(convert_to_rad)).to_bytes(4, byteorder="little", signed=False))
-        b.extend((int(interpolate)).to_bytes(4, byteorder="little", signed=False))
-        b.extend(tracking_file.encode("utf-8"))
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x03:
-            self.closeConnection()
-            raise Exception("Server Error saving tracking data")
+
+        self._send_command(
+            0x03,
+            volume,
+            int(save_as_matrix),
+            int(save_as_rows),
+            int(save_with_commas),
+            int(convert_to_cm),
+            int(convert_to_rad),
+            int(interpolate),
+            tracking_file,
+        )
 
     def loadFilters(self, camera, settings_file):
         """
@@ -182,22 +236,14 @@ class AutoscoperConnection:
         :type camera: int
         :param settings_file: The path to the filter settings to load
         :type settings_file: str
-        :raises Exception: If the filter settings file is not found, or If the server fails to load the filter settings
+        :raises AutoScoperServerError: If the server fails to load the filter settings
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Loading filter settings: {settings_file}")
         if not os.path.exists(settings_file):
-            self.closeConnection()
-            raise Exception("Filter settings file not found")
-        b = bytearray()
-        b.append(0x04)
-        b.extend(camera.to_bytes(4, byteorder="little", signed=False))
-        b.extend(settings_file.encode("utf-8"))
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x04:
-            self.closeConnection()
-            raise Exception("Server Error loading filter settings")
+            raise AutoscoperServerError(f"Filter settings not found: {settings_file}")
+        self._send_command(0x04, camera, settings_file)
 
     def setFrame(self, frame):
         """
@@ -205,18 +251,12 @@ class AutoscoperConnection:
 
         :param frame: The frame to be used for the next acquisition
         :type frame: int
-        :raises Exception: If the server fails to set the frame
+        :raises AutoScoperServerError: If the server fails to set the frame
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Setting frame: {frame}")
-        b = bytearray()
-        b.append(0x05)
-        b.extend(frame.to_bytes(4, byteorder="little", signed=False))
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x05:
-            self.closeConnection()
-            raise Exception("Server Error setting frame")
+        self._send_command(0x05, frame)
 
     def getPose(self, volume, frame):
         """
@@ -228,28 +268,19 @@ class AutoscoperConnection:
         :type frame: int
         :return: The pose of the volume at the specified frame
         :rtype: list[float]
-        :raises Exception: If the server fails to get the pose
+        :raises AutoScoperServerError: If the server fails to get the pose
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Getting pose for volume {volume} on frame {frame}")
-        b = bytearray()
-        b.append(0x06)
-        b.extend(volume.to_bytes(4, byteorder="little", signed=False))
-        b.extend(frame.to_bytes(4, byteorder="little", signed=False))
-        self.socket.sendall(b)
-        data = self.wait_for_server()
-        print(data[0])
-        if data[0] != 0x06:
-            self.closeConnection()
-            raise Exception("Server Error getting pose")
-        data = bytearray(data)
+        response = self._send_command(0x06, volume, frame)
         return [
-            struct.unpack("d", data[1:9])[0],
-            struct.unpack("d", data[9:17])[0],
-            struct.unpack("d", data[17:25])[0],
-            struct.unpack("d", data[25:33])[0],
-            struct.unpack("d", data[33:41])[0],
-            struct.unpack("d", data[41:49])[0],
+            struct.unpack("d", response[1:9])[0],
+            struct.unpack("d", response[9:17])[0],
+            struct.unpack("d", response[17:25])[0],
+            struct.unpack("d", response[25:33])[0],
+            struct.unpack("d", response[33:41])[0],
+            struct.unpack("d", response[41:49])[0],
         ]
 
     def setPose(self, volume, frame, pose):
@@ -262,25 +293,12 @@ class AutoscoperConnection:
         :type frame: int
         :param pose: The pose to set the volume to
         :type pose: list[float]
-        :raises Exception: If the server fails to set the pose
+        :raises AutoScoperServerError: If the server fails to set the pose
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Setting pose {pose} for volume {volume} on frame {frame}")
-        b = bytearray()
-        b.append(0x07)
-        b.extend(volume.to_bytes(4, byteorder="little", signed=False))
-        b.extend(frame.to_bytes(4, byteorder="little", signed=False))
-        b.extend(struct.pack("d", pose[0]))
-        b.extend(struct.pack("d", pose[1]))
-        b.extend(struct.pack("d", pose[2]))
-        b.extend(struct.pack("d", pose[3]))
-        b.extend(struct.pack("d", pose[4]))
-        b.extend(struct.pack("d", pose[5]))
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x07:
-            self.closeConnection()
-            raise Exception("Server Error setting pose")
+        self._send_command(0x07, volume, frame, *pose)
 
     def getNCC(self, volume, pose):
         """
@@ -292,28 +310,15 @@ class AutoscoperConnection:
         :type pose: list[float]
         :return: The NCC of the volume at the specified pose
         :rtype: list[float]
-        :raises Exception: If the server fails to get the NCC
+        :raises AutoScoperServerError: If the server fails to get the NCC
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Getting NCC for volume {volume} on pose {pose}")
-        b = bytearray()
-        b.append(0x08)
-        b.extend(volume.to_bytes(4, byteorder="little", signed=False))
-        b.extend(struct.pack("d", pose[0]))
-        b.extend(struct.pack("d", pose[1]))
-        b.extend(struct.pack("d", pose[2]))
-        b.extend(struct.pack("d", pose[3]))
-        b.extend(struct.pack("d", pose[4]))
-        b.extend(struct.pack("d", pose[5]))
-        self.socket.sendall(b)
-        data = self.wait_for_server()
-        if data[0] != 0x08:
-            self.closeConnection()
-            raise Exception("Server Error getting NCC")
-        data = bytearray(data)
+        response = self._send_command(0x08, volume, *pose)
         ncc = []
         for i in range(0, 2):
-            val = data[2 + (i) * 8 : 10 + (i) * 8]
+            val = response[2 + (i) * 8 : 10 + (i) * 8]
             ncc.append(struct.unpack("d", val)[0])
         return ncc
 
@@ -323,18 +328,12 @@ class AutoscoperConnection:
 
         :param threshold: The background threshold
         :type threshold: float
-        :raises Exception: If the server fails to set the background threshold
+        :raises AutoScoperServerError: If the server fails to set the background threshold
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(f"Setting background threshold: {threshold}")
-        b = bytearray()
-        b.append(0x09)
-        b.extend(struct.pack("d", threshold))
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x09:
-            self.closeConnection()
-            raise Exception("Server Error setting background threshold")
+        self._send_command(0x09, threshold)
 
     def getImageCropped(self, volume, camera, pose):
         """
@@ -354,25 +353,10 @@ class AutoscoperConnection:
             print(
                 f"Getting image for volume {volume} on pose {pose} from camera {camera}"
             )
-        b = bytearray()
-        b.append(0x0A)
-        b.extend(volume.to_bytes(4, byteorder="little", signed=False))
-        b.extend(camera.to_bytes(4, byteorder="little", signed=False))
-        b.extend(struct.pack("d", pose[0]))
-        b.extend(struct.pack("d", pose[1]))
-        b.extend(struct.pack("d", pose[2]))
-        b.extend(struct.pack("d", pose[3]))
-        b.extend(struct.pack("d", pose[4]))
-        b.extend(struct.pack("d", pose[5]))
-        self.socket.sendall(b)
-        data = self.wait_for_server()
-        if data[0] != 0x0A:
-            self.closeConnection()
-            raise Exception("Server Error getting image")
-        data = bytearray(data)
-        width = struct.unpack("i", data[1:5])[0]
-        height = struct.unpack("i", data[5:9])[0]
-        img_data = data[9:]
+        response = self._send_command(0x0A, volume, camera, *pose)  # 10
+        width = struct.unpack("i", response[1:5])[0]
+        height = struct.unpack("i", response[5:9])[0]
+        img_data = response[9:]
         return [width, height, img_data]
 
     def optimizeFrame(
@@ -411,7 +395,8 @@ class AutoscoperConnection:
         :type opt_method: int
         :param cf_model: The cost function model to use, 0 for NCC (Bone Models), 1 for Sum of Absolute Differences (Implant Models)
         :type cf_model: int
-        :raises Exception: If the server fails to optimize the frame
+        :raises AutoscoperServerError: If the server fails to optimize the frame
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if opt_method not in [0, 1]:
             raise Exception("Invalid optimization method")
@@ -419,49 +404,43 @@ class AutoscoperConnection:
             raise Exception("Invalid cost function model")
         if self.verbose:
             print(f"Optimizing volume {volume} on frame {frame}")
-        b = bytearray()
-        b.append(0x0B)
-        b.extend(volume.to_bytes(4, byteorder="little", signed=False))
-        b.extend(frame.to_bytes(4, byteorder="little", signed=False))
-        b.extend(repeats.to_bytes(4, byteorder="little", signed=False))
-        b.extend(max_itr.to_bytes(4, byteorder="little", signed=False))
-        b.extend(struct.pack("d", min_lim))
-        b.extend(struct.pack("d", max_lim))
-        b.extend(max_stall_itr.to_bytes(4, byteorder="little", signed=False))
-        b.extend(dframe.to_bytes(4, byteorder="little", signed=False))
-        b.extend(opt_method.to_bytes(4, byteorder="little", signed=False))
-        b.extend(cf_model.to_bytes(4, byteorder="little", signed=False))
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x0B:
-            self.closeConnection()
-            raise Exception("Server Error optimizing frame")
+        self._send_command(
+            0x0B,  # 11
+            volume,
+            frame,
+            repeats,
+            max_itr,
+            min_lim,
+            max_lim,
+            max_stall_itr,
+            dframe,
+            opt_method,
+            cf_model,
+        )
 
     def saveFullDRR(self):
         """
         Save the full DRR.
 
-        :raises Exception: If the server fails to save the full DRR
+        :raises AutoscoperServerError: If the server fails to save the full DRR
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
-        b = bytearray()
-        b.append(0x0C)
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if int.from_bytes(res, byteorder="little", signed=False) != 0x0C:
-            self.closeConnection()
-            raise Exception("Server Error saving full DRR")
+        self._send_command(0x0C)  # 12
 
     def closeConnection(self):
         """
         Close the connection to the server.
 
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         b = bytearray()
         # convert 13 to bytes
         b.append(0x0D)
-        self.socket.sendall(b)
+        try:
+            self.socket.sendall(b)
+        except socket.error as e:
+            raise AutoscoperConnectionError("Connection to server lost") from e
         self.socket.close()
-        self.is_connected = False
 
     def trackingDialog(
         self,
@@ -500,6 +479,12 @@ class AutoscoperConnection:
         :type max_lim: float
         :param max_stall_itr: The maximum number of iterations to stall
         :type max_stall_itr: int
+        :param opt_method: The optimization method to use, 0 for Particle Swarm, 1 for Downhill Simplex
+        :type opt_method: int
+        :param cf_model: The cost function model to use, 0 for NCC (Bone Models), 1 for Sum of Absolute Differences (Implant Models)
+        :type cf_model: int
+        :raises AutoscoperServerError: If the server fails to track the volume
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
         if self.verbose:
             print(
@@ -529,15 +514,12 @@ class AutoscoperConnection:
 
         :return: The number of volumes in the scene
         :rtype: int
+        :raises AutoscoperServerError: If the server fails to get the number of volumes
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
-        b = bytearray()
-        b.append(0x0E)
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if res[0] != 0x0E:
-            self.closeConnection()
-            raise Exception("Server Error getting number of volumes")
-        return int.from_bytes(res[1:], byteorder="little", signed=False)
+        response = self._send_command(0x0E)  # 14
+        num_volume = struct.unpack("i", response[1:])[0]
+        return num_volume
 
     def getNumFrames(self):
         """
@@ -545,13 +527,9 @@ class AutoscoperConnection:
 
         :return: The number of frames in the scene
         :rtype: int
+        :raises AutoscoperServerError: If the server fails to get the number of frames
+        :raises AutoscoperConnectionError: If the connection to the server is lost
         """
-        b = bytearray()
-        b.append(0x0F)
-        self.socket.sendall(b)
-        res = self.wait_for_server()
-        if res[0] != 0x0F:
-            self.closeConnection()
-            raise Exception("Server Error getting number of frames")
-        # the rest of the bytes are the number of frames
-        return int.from_bytes(res[1:], byteorder="little", signed=True)
+        response = self._send_command(0x0F)  # 15
+        num_frames = struct.unpack("i", response[1:])[0]
+        return num_frames
